@@ -22,8 +22,14 @@ SOFTWARE.
 
 int setup(void);
 void setup_gpio(int gpio, int direction, int pud);
+int gpio_function(int gpio);
 void output_gpio(int gpio, int value);
 int input_gpio(int gpio);
+void set_rising_event(int gpio, int enable);
+void set_falling_event(int gpio, int enable);
+void set_high_event(int gpio, int enable);
+void set_low_event(int gpio, int enable);
+int event_detected(int gpio);
 void cleanup(void);
 
 #define SETUP_OK          0
@@ -31,8 +37,9 @@ void cleanup(void);
 #define SETUP_MALLOC_FAIL 2
 #define SETUP_MMAP_FAIL   3
 
-#define INPUT  1
-#define OUTPUT 0
+#define INPUT  1 // is really 0 for control register!
+#define OUTPUT 0 // is really 1 for control register!
+#define ALT0   4
 
 #define HIGH 1
 #define LOW  0
